///////////////////////////////////////////////////////////////////////////////
// File: ServerMain.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2016 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/StringConverter.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/GlobalErrorHandler.h"
#include "JetByteTools/CoreTools/ThreadedCallbackTimerQueue.h"
#include "JetByteTools/CoreTools/DebugTrace.h"

#include "JetByteTools/IOTools/AsyncFileLog.h"

#include "JetByteTools/SocketTools/FullAddress.h"
#include "JetByteTools/SocketTools/AddressTypeIPv4.h"
#include "JetByteTools/SocketTools/DatagramSocketAllocator.h"
#include "JetByteTools/SocketTools/ServerCollection.h"
#include "JetByteTools/SocketTools/SequencedStreamSocketAllocator.h"

#include "JetByteTools/HTTPTools/HTTPHeaders.h"
#include "JetByteTools/HTTPTools/HTTPResourceManager.h"
#include "JetByteTools/HTTPTools/HTTPSocketServer.h"

#include "JetByteTools/StreamingTools/HLSFileResource.h"

#include "ServerCommon/IOPool.h"
#include "ServerCommon/BufferAllocator.h"
#include "ServerCommon/SimpleServerShutdownHandler.h"

#pragma hdrstop

#include <iostream>

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Milliseconds;
using JetByteTools::Seconds;

using JetByteTools::Core::_tstring;
using JetByteTools::Core::ToString;
using JetByteTools::Core::CException;
using JetByteTools::Core::OutputEx;
using JetByteTools::Core::CDebugTrace;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::CGlobalErrorHandler;
using JetByteTools::Core::CThreadedCallbackTimerQueue;

using JetByteTools::IO::CAsyncFileLog;
using JetByteTools::IO::CSmartBuffer;

using JetByteTools::Socket::CFullAddress;
using JetByteTools::Socket::CAddressTypeIPv4;
using JetByteTools::Socket::CAddressType;
using JetByteTools::Socket::CDatagramSocketAllocator;
using JetByteTools::Socket::CFlowControlStreamSocketConnectionFilter;
using JetByteTools::Socket::SocketBufferSize;
using JetByteTools::Socket::DefaultSocketBufferSize;
using JetByteTools::Socket::CServerCollection;
using JetByteTools::Socket::CSequencedStreamSocketAllocator;

using JetByteTools::Streaming::CHLSFileResource;

using JetByteTools::HTTP::CHTTPHeaders;
using JetByteTools::HTTP::CHTTPResourceManager;
using JetByteTools::HTTP::CSmartHTTPResource;
using JetByteTools::HTTP::CHTTPSocketServer;

using std::cerr;
using std::endl;

///////////////////////////////////////////////////////////////////////////////
// Program entry point
///////////////////////////////////////////////////////////////////////////////

int main(int /*argc*/, char * /*argv[ ]*/)
{
   CGlobalErrorHandler errorHandler;

   int result = 0;

   try
   {
      CSimpleServerShutdownHandler shutdownHandler;

      CAsyncFileLog log(_T("HLSServer.log"));

      CDebugTrace::LogInstaller logInstaller(log);

      try
      {
         const bool displayDebug = false;

         CIOPool socketIoPool(
            2,                            // number of threads (0 = 2 x processors)
            displayDebug);

         socketIoPool.Start();

         CIOPool fileIoPool(
            2,                            // number of threads (0 = 2 x processors)
            displayDebug);

         fileIoPool.Start();

         CSequencedStreamSocketAllocator socketAllocator(
            10,                              // max number of sockets to keep in the pool
            10,                              // max number of buffers pooled in the socket
            10);                             // max number of consecutive socket operations on this socket

         CThreadedCallbackTimerQueue timerQueue;

         const Seconds sessionInactivityTimeout = 30;

         CBufferAllocator bufferAllocator(
            40960,                         // buffer size
            100,                           // max number of buffers to keep in the pool
            displayDebug);

         const _tstring baseUri(_T("/"));

         CHTTPResourceManager resourceManager(
            baseUri);

         CHTTPHeaders standardHeaders;

         standardHeaders.Add("Cache-Control: private, max-age=0, no-cache, no-store");
         standardHeaders.Add("Pragma: no-cache");
         standardHeaders.Add("Expires: Mon, 29 Apr 2013 21:44:55 GMT");               // always expired
         standardHeaders.Add("Access-Control-Allow-Origin: *");

         {
            CSmartHTTPResource resource(
               new CHLSFileResource(
                  _T("TestVideos\\Test1\\Index.m3u8"),
                  bufferAllocator,
                  fileIoPool,
                  timerQueue,
                  sessionInactivityTimeout,
                  standardHeaders));

            // could want to restrict this to a single m3u8 file (which the above implies)

            resourceManager.AddResource(_T("BipBop"), resource.GetRef());
         }

         const CAddressType addressPreference = CAddressTypeIPv4::instance;

         const size_t listenBacklog = 5;

         CServerCollection servers;

         {
            const CFullAddress address(
               _T("0.0.0.0"),               // address to listen on
               5050,                        // port to listen on
               addressPreference);

            // The following parameters affect how we handle flow control on the HTTP/HLS sockets.
            // Since we can be streaming video data on these connections we need to worry about what we do
            // if the players are not consuming data as fast as we're producing it.
            // Right now we 'pause' playback if the flow control list fills up and resume when it's empty again.
            // It may be better to resume sooner so that we've always got some data in flight.
            // These could all be made configurable.

            const CFlowControlStreamSocketConnectionFilter::BufferCount maxBuffers = 50;

            const CFlowControlStreamSocketConnectionFilter::PendingWriteCount maxPendingWrites = 50;

            const CFlowControlStreamSocketConnectionFilter::Percentage notificationPercentageStepSize = 10;

            const SocketBufferSize recvBufferSize = DefaultSocketBufferSize;

            const SocketBufferSize sendBufferSize = DefaultSocketBufferSize;

            servers.AddServer(
               new CHTTPSocketServer(
                  resourceManager,
                  maxBuffers,
                  maxPendingWrites,
                  notificationPercentageStepSize,
                  recvBufferSize,
                  sendBufferSize,
                  address,
                  listenBacklog,
                  socketIoPool,
                  socketAllocator,
                  bufferAllocator));
         }

         servers.Start();

         servers.StartAcceptingConnections();

         shutdownHandler.SetServerControl(servers);

         shutdownHandler.WaitForShutdownRequest();

         servers.WaitForShutdownToComplete();

         resourceManager.WaitForShutdownToComplete();

         socketIoPool.WaitForShutdownToComplete();

         fileIoPool.WaitForShutdownToComplete();

         socketAllocator.ReleaseSockets();

         bufferAllocator.Flush();
      }
      catch(const CException &e)
      {
         OutputEx(_T("Exception: ") + e.GetDetails());

         result = 1;
      }
      catch(...)
      {
         OutputEx(_T("Unexpected exception"));

         result = 1;
      }
   }
   catch(const CException &e)
   {
      cerr << "Exception: " << e.GetDetailsA() << endl;

      result = 1;
   }
   catch(...)
   {
      cerr << "Unexpected exception" << endl;

      result = 1;
   }

   return result;
}

///////////////////////////////////////////////////////////////////////////////
// End of file: ServerMain.cpp
///////////////////////////////////////////////////////////////////////////////
