///////////////////////////////////////////////////////////////////////////////
// File: ServerMain.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2014 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "SocketServer.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/StringConverter.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/GlobalErrorHandler.h"
#include "JetByteTools/CoreTools/ThreadedCallbackTimerQueue.h"
#include "JetByteTools/CoreTools/DebugTrace.h"

#include "JetByteTools/Win32Tools/SystemUtils.h"

#include "JetByteTools/IOTools/AsyncFileLog.h"

#include "JetByteTools/SocketTools/FullAddress.h"
#include "JetByteTools/SocketTools/AddressTypeIPv4.h"
#include "JetByteTools/SocketTools/DatagramSocketAllocator.h"
#include "JetByteTools/SocketTools/ServerCollection.h"
#include "JetByteTools/SocketTools/SequencedStreamSocketAllocator.h"

#include "JetByteTools/StreamingTools/RTSPResourceManager.h"
#include "JetByteTools/StreamingTools/RTSPH264FileResource.h"
#include "JetByteTools/StreamingTools/RTSPMP3FileResource.h"
#include "JetByteTools/StreamingTools/RTSPMultipleFileResource.h"
#include "JetByteTools/StreamingTools/UDPConnectionCreator.h"

#include "ServerCommon/IOPool.h"
#include "ServerCommon/BufferAllocator.h"
#include "ServerCommon/SimpleServerShutdownHandler.h"

#pragma hdrstop

#include <iostream>

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Milliseconds;
using JetByteTools::Seconds;

using JetByteTools::Core::_tstring;
using JetByteTools::Core::ToString;
using JetByteTools::Core::CException;
using JetByteTools::Core::OutputEx;
using JetByteTools::Core::CDebugTrace;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::GetStringLength;
using JetByteTools::Core::CGlobalErrorHandler;
using JetByteTools::Core::CThreadedCallbackTimerQueue;

using JetByteTools::Win32::GetComputerName;
using JetByteTools::Win32::GetUserName;

using JetByteTools::IO::CAsyncFileLog;
using JetByteTools::IO::CSmartBuffer;

using JetByteTools::Socket::CFullAddress;
using JetByteTools::Socket::CAddressTypeIPv4;
using JetByteTools::Socket::CAddressType;
using JetByteTools::Socket::CDatagramSocketAllocator;
using JetByteTools::Socket::CFlowControlStreamSocketConnectionFilter;
using JetByteTools::Socket::SocketBufferSize;
using JetByteTools::Socket::DefaultSocketBufferSize;
using JetByteTools::Socket::CServerCollection;
using JetByteTools::Socket::CSequencedStreamSocketAllocator;

using JetByteTools::Streaming::CRTSPResourceManager;
using JetByteTools::Streaming::CRTSPH264FileResource;
using JetByteTools::Streaming::CRTSPMP3FileResource;
using JetByteTools::Streaming::CRTSPMultipleFileResource;
using JetByteTools::Streaming::CSmartRTSPResource;
using JetByteTools::Streaming::CUDPConnectionCreator;

using std::cerr;
using std::endl;

///////////////////////////////////////////////////////////////////////////////
// Program entry point
///////////////////////////////////////////////////////////////////////////////

int main(int /*argc*/, char * /*argv[ ]*/)
{
   CGlobalErrorHandler errorHandler;

   auto result = 0;

   try
   {
      CSimpleServerShutdownHandler shutdownHandler;

      CAsyncFileLog log(_T("RTSPServer.log"));

      CDebugTrace::LogInstaller logInstaller(log);

      try
      {
         const auto displayDebug = false;

         CIOPool socketIoPool(
            2,                            // number of threads (0 = 2 x processors)
            displayDebug);

         socketIoPool.Start();

         CIOPool fileIoPool(
            2,                            // number of threads (0 = 2 x processors)
            displayDebug);

         fileIoPool.Start();

         CSequencedStreamSocketAllocator socketAllocator(
            10,                              // max number of sockets to keep in the pool
            10,                              // max number of buffers pooled in the socket
            10);                             // max number of consecutive socket operations on this socket

         CBufferAllocator bufferAllocator(
            40960,                         // buffer size
            100,                           // max number of buffers to keep in the pool
            displayDebug);

         const _tstring baseUri(_T("/"));

         const DWORD framesToCache = 24 * 60;      // 20 seconds of video

         const DWORD mtu = 1448;

         CDatagramSocketAllocator udpSocketAllocator(
            10,                              // max number of sockets to keep in the pool
            10,                              // max number of buffers pooled in the socket
            10);                             // max number of consecutive socket operations on this socket

         CUDPConnectionCreator udpConnectionCreator(
            socketIoPool,
            udpSocketAllocator,
            bufferAllocator);

         CThreadedCallbackTimerQueue timerQueue;

         const Seconds sessionInactivityTimeout = 30;

         const auto defaultCName = GetUserName() + _T("@") + GetComputerName();

         CRTSPResourceManager resourceManager(
            defaultCName,
            baseUri,
            framesToCache,                         // should be in seconds?
            timerQueue,
            sessionInactivityTimeout,
            bufferAllocator,
            udpConnectionCreator,
            mtu);

         {
            CSmartRTSPResource resource(
               new CRTSPH264FileResource(
               _T("Big Buck Bunny"),
               _T("TestVideos\\big_buck_bunny_480p_h264.h264"),
               bufferAllocator,
               fileIoPool));

            resourceManager.AddResource(_T("BBB"), resource.GetRef());
         }

         const CFlowControlStreamSocketConnectionFilter::BufferCount maxBuffers = 10;

         const CFlowControlStreamSocketConnectionFilter::PendingWriteCount maxPendingWrites = 10;

         const CFlowControlStreamSocketConnectionFilter::Percentage notificationPercentageStepSize = 25;

         const SocketBufferSize recvBufferSize = DefaultSocketBufferSize;

         const SocketBufferSize sendBufferSize = DefaultSocketBufferSize;

         const auto addressPreference = CAddressTypeIPv4::instance;

         const size_t listenBacklog = 5;

         const auto terminateActiveSessionsOnConnectionClose = false;

         CServerCollection servers;

         {
            const CFullAddress address(
               _T("0.0.0.0"),               // address to listen on
               8554,                        // port to listen on
               addressPreference);

            servers.AddServer(
               new CSocketServer(
                  CRTSPResourceManager::SupportsAllTransports,
                  resourceManager,
                  terminateActiveSessionsOnConnectionClose,
                  maxBuffers,
                  maxPendingWrites,
                  notificationPercentageStepSize,
                  recvBufferSize,
                  sendBufferSize,
                  address,
                  listenBacklog,
                  socketIoPool,
                  socketAllocator,
                  bufferAllocator));
         }

         {
            const CFullAddress address(
               _T("0.0.0.0"),               // address to listen on
               8555,                        // port to listen on
               addressPreference);

            servers.AddServer(
               new CSocketServer(
                  CRTSPResourceManager::SupportsUDPTransport,
                  resourceManager,
                  terminateActiveSessionsOnConnectionClose,
                  maxBuffers,
                  maxPendingWrites,
                  notificationPercentageStepSize,
                  recvBufferSize,
                  sendBufferSize,
                  address,
                  listenBacklog,
                  socketIoPool,
                  socketAllocator,
                  bufferAllocator));
         }

         {
            const CFullAddress address(
               _T("0.0.0.0"),               // address to listen on
               8556,                        // port to listen on
               addressPreference);

            servers.AddServer(
               new CSocketServer(
                  CRTSPResourceManager::SupportsTCPTransport,
                  resourceManager,
                  terminateActiveSessionsOnConnectionClose,
                  maxBuffers,
                  maxPendingWrites,
                  notificationPercentageStepSize,
                  recvBufferSize,
                  sendBufferSize,
                  address,
                  listenBacklog,
                  socketIoPool,
                  socketAllocator,
                  bufferAllocator));
         }

         servers.Start();

         servers.StartAcceptingConnections();

         shutdownHandler.SetServerControl(servers);

         shutdownHandler.WaitForShutdownRequest();

         servers.WaitForShutdownToComplete();

         resourceManager.WaitForShutdownToComplete();

         udpConnectionCreator.WaitForShutdownToComplete();

         socketIoPool.WaitForShutdownToComplete();

         fileIoPool.WaitForShutdownToComplete();

         socketAllocator.ReleaseSockets();

         bufferAllocator.Flush();
      }
      catch(const CException &e)
      {
         OutputEx(_T("Exception: ") + e.GetDetails());

         result = 1;
      }
      catch(...)
      {
         OutputEx(_T("Unexpected exception"));

         result = 1;
      }
   }
   catch(const CException &e)
   {
      cerr << "Exception: " << e.GetDetailsA() << endl;

      result = 1;
   }
   catch(...)
   {
      cerr << "Unexpected exception" << endl;

      result = 1;
   }

   return result;
}

///////////////////////////////////////////////////////////////////////////////
// End of file: ServerMain.cpp
///////////////////////////////////////////////////////////////////////////////
